/*  Eside 3D library, version 5.
    Copyright (C) 2002 Iker Jamardo Zugaza <ijamardo@eside.deusto.es>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    This library has been slightly modified by Grzegorz Adam Hankiewicz
    <gradha@terra.es> and is therefor not exactly the original, which you
    should be able to get from Iker by email.
*/

#ifndef _E3D_CMATRIX4D_H_
#define _E3D_CMATRIX4D_H_

#pragma warning (disable: 4786 4503)

#include "E3D_CObject.h"
#include "E3D_CVector4D.h"
#include "E3D_CVector3D.h"
#include "E3D_CBoundingBox.h"

// ============================================================================
// The class that represents a 4x4 (4D) matrix with all the important operations
// that needs to perform. The 4rd coordinate is the homogeneous coordinate.
// This matrix post-multiplies a vector.
// ============================================================================
class CMatrix4D: public CObject
{
public:
// From CObject
	// ============================================================================
	// An operator to cast any object to a string. Polimorph it to give it a new
	// behaviour that fits your needs.
	// ============================================================================
	virtual operator string() const;

	// ============================================================================
	// An operator to compare the equality of two objects. Polimorph it to give
	// a new behaviour that fits your needs.
	// ============================================================================
	virtual bool operator==(const CObject& obj) const;

// Own
	// ============================================================================
	// Add each component of the matrices and store the result inside a new matrix.
	// ============================================================================
	CMatrix4D operator+(const CMatrix4D& m) const;

	// ============================================================================
	// Substract each component of the matrices and store the result inside a new matrix
	// ============================================================================
	CMatrix4D operator-(const CMatrix4D& m) const;

	// ============================================================================
	// ============================================================================
	CBoundingBox operator*(const CBoundingBox& bb) const;

	// ============================================================================
	// Multiply two matrices and store the result inside a new matrix.
	// ============================================================================
	CMatrix4D operator*(const CMatrix4D& m) const;

	// ============================================================================
	// Multiply this matrix by a vector and get a new vector. (3D)
	// ============================================================================
	CVector3D operator*(const CVector3D& v) const;

	// ============================================================================
	// Multiply this matrix by a vector and get a new vector. (4D)
	// ============================================================================
	CVector4D operator*(const CVector4D& v) const;

	// ============================================================================
	// Multiply each component of the matrix by a factor getting a new matrix.
	// ============================================================================
	CMatrix4D operator*(float factor) const;

	// ============================================================================
	// Multiply this matrix by each 3D vector stored inside the parameter
	// container and return a new container with the result
	// ============================================================================
	CVector3DVector operator*(const CVector3DVector& vectors) const;

	// ============================================================================
	// Apply all the transformations by matrix concatenation.
	// ============================================================================
	void translateRotateScale(const CVector3D& t, const CVector3D& r, const CVector3D& s);

	// ============================================================================
	// Make this matrix to be a perspective projection matrix.
	// ============================================================================
	void perspectiveProjection(float l, float r, float b, float t, float zn, float zf);

	// ============================================================================
	// Make this matrix to be a orthographic projection matrix.
	// ============================================================================
	void orthographicProjection(float l, float r, float b, float t, float zn, float zf);

	// ============================================================================
	// Get the inverse of a given matrix.
	// ============================================================================
	static CMatrix4D inverse(const CMatrix4D& m);

	// ============================================================================
	// Make this matrix to be identity.
	// ============================================================================
	void identity();

	// ============================================================================
	// Make this matrix to be zero.
	// ============================================================================
	void zero();

	// ============================================================================
	// Store this matrix inside a float array.
	// ============================================================================
	bool toArray(float* array) const;

	// ============================================================================
	// Get this matrix components values from a float array.
	// ============================================================================
	void fromArray(float* array);

  CVector4D x, y, z, w;
};

// ============================================================================
// An operator to let change the order for matrix and scalar product.
// ============================================================================
CMatrix4D operator*(float factor, const CMatrix4D& m);

#endif