/*  Eside 3D library, version 5.
    Copyright (C) 2002 Iker Jamardo Zugaza <ijamardo@eside.deusto.es>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    This library has been slightly modified by Grzegorz Adam Hankiewicz
    <gradha@terra.es> and is therefor not exactly the original, which you
    should be able to get from Iker by email.
*/

#ifndef _E3D_CVECTOR3D_H_
#define _E3D_CVECTOR3D_H_

#include "E3D_CObject.h"

#include <vector>
using std::vector;
using std::bad_cast;

class CVector4D;

// ============================================================================
// The class that represents a 3D vector with all the important operations
// that needs to perform.
// ============================================================================
class CVector3D: public CObject
{
public:
// From CObject
	// ============================================================================
	// An operator to cast any object to a string. Polimorph it to give it a new
	// behaviour that fits your needs.
	// ============================================================================
	virtual operator string() const;

	// ============================================================================
	// An operator to compare the equality of two objects. Polimorph it to give
	// a new behaviour that fits your needs.
	// ============================================================================
	virtual bool operator==(const CObject& obj) const;

	// ============================================================================
	// A function to take all the data that an object needs from an string.
	// ============================================================================
	virtual bool fromString(const string& from);

// Own
	// ============================================================================
	// Constructor. Initializes all the x, y, z real values to 0.
	// ============================================================================
	CVector3D();

	// ============================================================================
	// Constructor. Initializes all the x, y, z values to parameter values.
	// ============================================================================
	CVector3D(float x, float y, float z);

	// ============================================================================
	// Operator to add two vectors. Just add each component of each vector. The 
	// result is a new vector (so operands don't change).
	// ============================================================================
	CVector3D operator+(const CVector3D& v) const;

	// ============================================================================
	// Operator to add two vectors. The result is stored inside the this vector (the
	// first operand).
	// ============================================================================
	void operator+=(const CVector3D& v);

	// ============================================================================
	// Operator to substract two vectors. Just substract each component of each vector. 
	// The result is a new vector (so operands don't change).
	// ============================================================================
	CVector3D operator-(const CVector3D& v) const;

	// ============================================================================
	// Operator to substract two vectors. The result is stored inside the this vector
	// (the first operand).
	// ============================================================================
	void operator-=(const CVector3D& v);

	// ============================================================================
	// This operator negates the vector components.
	// ============================================================================
	CVector3D operator-();

	// ============================================================================
	// Multiplies each component of the two vector operands to obtain a new vector.
	// ============================================================================
	CVector3D operator*(const CVector3D& v) const;

	// ============================================================================
	// Multiplies each component with a factor.
	// ============================================================================
	CVector3D operator*(float factor) const;

	// ============================================================================
	// Multiplies each component of the two vector operands and the vector result 
	// is stored inside the first.
	// ============================================================================
	void operator*=(const CVector3D& v);

	// ============================================================================
	// Multiplies each component of the vector by the factor and the result is stored
	// inside the vector.
	// ============================================================================
	void operator*=(float factor);

	// ============================================================================
	// Divides each components of the two vector operands and the result is a new vector.
	// ============================================================================
	CVector3D operator/(const CVector3D& v) const;

	// ============================================================================
	// Divides each component of the vector by a factor and the result is a new vector.
	// ============================================================================
	CVector3D operator/(float factor) const;

	// ============================================================================
	// Divides each components of the two vector operands and the result is stored
	// inside the first operand.
	// ============================================================================
	void operator/=(const CVector3D& v);

	// ============================================================================
	// Divides each components of the vector by a factor and the result is stored
	// inside the first operand vector.
	// ============================================================================
	void operator/=(float factor);

	// ============================================================================
	// The cross product.
	// ============================================================================
	CVector3D operator%(const CVector3D& v) const;

	// ============================================================================
	// The cross product.
	// ============================================================================
	CVector4D operator%(const CVector4D& v) const;

	// ============================================================================
	// The dot product.
	// ============================================================================
	float operator^(const CVector4D& v) const;

	// ============================================================================
	// The dot product.
	// ============================================================================
	float operator^(const CVector3D& v) const;

	// ============================================================================
	// Operator to be able to cast from a 3D vector to a 4D vector
	// ============================================================================
	operator CVector4D() const;

	// ============================================================================
	// Operator to be able to asign a 3d and a 4d vector.
	// ============================================================================
	CVector3D operator=(const CVector4D& v);

  float x, y, z;
};

// ============================================================================
// Operators to be able to multiply a scalar by a vector (change the order)
// ============================================================================
CVector3D operator*(float factor, const CVector3D& v);

// ============================================================================
// Operators to be able to divide a scalar by a vector (change the order)
// ============================================================================
CVector3D operator/(float factor, const CVector3D& v);

typedef vector<CVector3D> CVector3DVector;
typedef CVector3DVector::iterator CVector3DVectorIterator;

#endif
