/*  Eside 3D library, version 5.
    Copyright (C) 2002 Iker Jamardo Zugaza <ijamardo@eside.deusto.es>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    This library has been slightly modified by Grzegorz Adam Hankiewicz
    <gradha@terra.es> and is therefor not exactly the original, which you
    should be able to get from Iker by email.
*/

#ifndef _E3D_CVECTOR4D_H_
#define _E3D_CVECTOR4D_H_

#include "E3D_CObject.h"

class CVector3D;

// ============================================================================
// The class that represents a 4D vector with all the important operations
// that needs to perform. A 4D vector is a 3D vector with the homogeneous coordinate.
// ============================================================================
class CVector4D: public CObject
{
public:
// From CObject
	// ============================================================================
	// An operator to cast any object to a string. Polimorph it to give it a new
	// behaviour that fits your needs.
	// ============================================================================
	virtual operator string() const;

	// ============================================================================
	// An operator to compare the equality of two objects. Polimorph it to give
	// a new behaviour that fits your needs.
	// ============================================================================
	virtual bool operator==(const CObject& obj) const;

// Own
	// ============================================================================
	// Constructor. Initializes all the x, y, z, w real values to 0.
	// ============================================================================
	CVector4D();

	// ============================================================================
	// Constructor. Initializes all the x, y, z, w values to parameter values.
	// ============================================================================
	CVector4D(float x, float y, float z, float w);

	// ============================================================================
	// Operator to add two vectors. Just add each component of each vector. The 
	// result is a new vector (so operands don't change).
	// ============================================================================
	CVector4D operator+(const CVector4D& v) const;

	// ============================================================================
	// Operator to add two vectors. The result is stored inside the this vector (the
	// first operand).
	// ============================================================================
	void operator+=(const CVector4D& v);

	// ============================================================================
	// Operator to substract two vectors. Just substract each component of each vector. 
	// The result is a new vector (so operands don't change).
	// ============================================================================
	CVector4D operator-(const CVector4D& v) const;

	// ============================================================================
	// Operator to substract two vectors. The result is stored inside the this vector
	// (the first operand).
	// ============================================================================
	void operator-=(const CVector4D& v);

	// ============================================================================
	// This operator negates the vector components.
	// ============================================================================
	CVector4D operator-();

	// ============================================================================
	// Multiplies each component of the two vector operands to obtain a new vector.
	// ============================================================================
	CVector4D operator*(const CVector4D& v) const;

	// ============================================================================
	// Multiplies each component with a factor.
	// ============================================================================
	CVector4D operator*(float factor) const;

	// ============================================================================
	// Multiplies each component of the two vector operands and the vector result 
	// is stored inside the first.
	// ============================================================================
	void operator*=(const CVector4D& v);

	// ============================================================================
	// Multiplies each component of the vector by the factor and the result is stored
	// inside the vector.
	// ============================================================================
	void operator*=(float factor);

	// ============================================================================
	// Divides each components of the two vector operands and the result is a new vector.
	// ============================================================================
	CVector4D operator/(const CVector4D& v) const;

	// ============================================================================
	// Divides each component of the vector by a factor and the result is a new vector.
	// ============================================================================
	CVector4D operator/(float factor) const;

	// ============================================================================
	// Divides each components of the two vector operands and the result is stored
	// inside the first operand.
	// ============================================================================
	void operator/=(const CVector4D& v);

	// ============================================================================
	// Divides each components of the vector by a factor and the result is stored
	// inside the first operand vector.
	// ============================================================================
	void operator/=(float factor);

	// ============================================================================
	// The cross product.
	// ============================================================================
	CVector4D operator%(const CVector4D& v) const;

	// ============================================================================
	// The cross product.
	// ============================================================================
	CVector4D operator%(const CVector3D& v) const;

	// ============================================================================
	// The dot product.
	// ============================================================================
	float operator^(const CVector4D& v) const;

	// ============================================================================
	// The dot product.
	// ============================================================================
	float operator^(const CVector3D& v) const;

	// ============================================================================
	// Operator to be able to cast from a 4D vector to a 3D vector
	// ============================================================================
	operator CVector3D() const;

	// ============================================================================
	// Operator to be able to asign a 3d and a 4d vector.
	// ============================================================================
	CVector4D operator=(const CVector3D& v);

	float x, y, z, w;
};

// ============================================================================
// Operators to be able to multiply a scalar by a vector (change the order)
// ============================================================================
CVector4D operator*(float factor, const CVector4D& v);

// ============================================================================
// Operators to be able to divide a scalar by a vector (change the order)
// ============================================================================
CVector4D operator/(float factor, const CVector4D& v);

#endif