/*  Eside 3D library, version 5.
    Copyright (C) 2002 Iker Jamardo Zugaza <ijamardo@eside.deusto.es>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    This library has been slightly modified by Grzegorz Adam Hankiewicz
    <gradha@terra.es> and is therefor not exactly the original, which you
    should be able to get from Iker by email.
*/

#include "E3D_CBase.h"
#include "E3D_IReader.h"
#include "E3D_IWriter.h"

CBase::operator string() const
{
	return name;
}

bool CBase::operator==(const CObject& obj) const
{
	bool result = false;
	try
	{
		const CBase& base = static_cast<const CBase&>(obj);
		result = name == base.name;
	}
	catch(bad_cast e)
	{
	}
	return result;
}

CBase::CBase(): dirty(true),
								startTime(0),
								endTime(0),
								currentTime(0),
								lastTime(0)
{
	typeName = "CBase";
}

CBase::~CBase()
{
}

bool CBase::read(IReader* reader)
{
	if (reader == NULL) return false;
	bool result = true;
	if (result) result = reader->push("CBase");
	if (result) result = reader->read("name", name);
	if (result) result = reader->read("startTime", startTime);
	if (result) result = reader->read("endTime", endTime);
	if (result) result = reader->pop("CBase");
	return result;
}

bool CBase::write(IWriter* writer)
{
	if (writer == NULL) return false;
	bool result = true;
	if (result) result = writer->push("CBase");
	if (result) result = writer->write("name", name);
	if (result) result = writer->write("startTime", startTime);
	if (result) result = writer->write("endTime", endTime);
	if (result) result = writer->pop("CBase");
	return result;
}

void CBase::update(float time)
{
	lastTime = currentTime;
	currentTime = time;
}

float* CBase::getAttribute(const string& attributeName)
{
	return NULL;
}

bool CBase::connect(CGraph& graph)
{
	return true;
}

void CBase::setName(const string& name)
{
	this->name = name;
}

void CBase::setStartTime(float startTime)
{
	this->startTime = startTime;
}

void CBase::setEndTime(float endTime)
{
	this->endTime = endTime;
}

string CBase::getName()
{
	return name;
}

float CBase::getStartTime()
{
	return startTime;
}

float CBase::getEndTime()
{
	return endTime;
}

int CBase::getNumConnections()
{
	return connections.size();
}

CConnection* CBase::getConnection(int index)
{
	CConnection* connection = NULL;
	if (index >= 0 || index < connections.size())
	{
		connection = connections[index];
	}
	return connection;
}

bool CBase::addConnection(CBase* srcNode, const string& srcAttributeName, const string& dstAttributeName)
{
	// To create a connection we need the node and which of his attributes is going
	// to affect to which of my attributes. So lets find them all!
	if (srcNode == NULL)
	{
		return false;
	}

	float* srcAttribute = NULL; // The src attribute (src nodes attibute)
	srcAttribute = srcNode->getAttribute(srcAttributeName);
	if (srcAttribute == NULL)
	{
		return false;
	}

	float* dstAttribute = NULL; // The dst attribute (our attribute)
	dstAttribute = this->getAttribute(dstAttributeName);
	if (dstAttribute == NULL)
	{
		return false;
	}

	// Now that we have all we need, create the connection
	CConnection* newConnection = new CConnection();
	if (newConnection == NULL)
	{
		return false;
	}
	newConnection->srcNode = srcNode;
	newConnection->srcAttribute = srcAttribute;
	newConnection->srcNodeName = srcNode->getName();
	newConnection->srcAttributeName = srcAttributeName;
	newConnection->dstNode = this;
	newConnection->dstAttribute = dstAttribute;
	newConnection->dstNodeName = name;
	newConnection->dstAttributeName = dstAttributeName;

	// Is not a bad idea to check if the connection allready exists before adding it
	// Use the operator== inside CConnection to do so...
	pCConnectionVectorIterator it = connections.begin();
	for (; it != connections.end(); it++)
	{
		CConnection* c = *it;
		if (*c == *newConnection)
		{
			delete newConnection;
			return false;
		}
	}

	// After all the checks we can add the new connection safely
	connections.push_back(newConnection);
	
	return true;
}

void CBase::removeConnections()
{
	// Delete all the created connections
	pCConnectionVectorIterator it = connections.begin();
	for (; it != connections.begin(); it++)
	{
		delete *it;
	}
	connections.clear();
}

bool CBase::updateConnections(float time)
{
	bool changed = false;
	// Check every connections we have, update them and asign the value from the 
	// connected node attribute to our attribute.
	pCConnectionVectorIterator it = connections.begin();
	for (; it != connections.end(); it++)
	{
		CConnection* c = *it;
		if (c->isConnected())
		{
			changed = true;
			// Update the connected node
			c->srcNode->update(time);
			// Get his attribute and asign it to our attribute
			*(c->dstAttribute) = *(c->srcAttribute);
		}
	}
	return changed;
}
