/*  Eside 3D library, version 5.
    Copyright (C) 2002 Iker Jamardo Zugaza <ijamardo@eside.deusto.es>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    This library has been slightly modified by Grzegorz Adam Hankiewicz
    <gradha@terra.es> and is therefor not exactly the original, which you
    should be able to get from Iker by email.
*/

#include "E3D_CGraph.h"
#include "E3D_IReader.h"
#include "E3D_IWriter.h"
#include "E3D_CTransform.h"

bool CGraph::read(IReader* reader)
{
	if (reader == NULL) return false;
	int i;
	bool result = CHierarchyBase::read(reader);
	end();
	if (result) result = reader->push("CGraph");

	if (result) result = reader->push("Nodes");
	if (result)
	{
		CBase* node = NULL;
		int numNodes;
		result = reader->read("Num", numNodes);
		for (i=0; i<numNodes && result; i++)
		{
			if (result = reader->push("Node"))
			{
				string typeName;
				if (result = reader->read("Type", typeName))
				{
					CFactoryObject* factory = CFactoryObject::getInstance();
					CObject* obj = factory->create(typeName);
					if (obj != NULL)
					{
						node = (CBase*)obj;
						if (result = node->read(reader))
						{
							result = addNode(node);
						}
					}
				}
			}
			if (result) result = reader->pop("Node");
		}
	}
	if (result) result = reader->pop("Nodes");

	if (result) result = reader->push("Hierarchy");
	if (result)
	{
		int numHierarchies;
		CBase* node1 = NULL;
		CBase* node2 = NULL;
		CHierarchyBase* parent = NULL;
		CHierarchyBase* child = NULL;
		string parentName, childName;
		result = reader->read("Num", numHierarchies);
		for (i=0; i<numHierarchies && result; i++)
		{
			result = reader->read("Parent", parentName);
			if (result) result = reader->read("Child", childName);
			if (result)
			{
				node1 = getNode(parentName);
				node2 = getNode(childName);
				if (node1 != NULL && node2 != NULL)
				{
					parent = (CHierarchyBase*)node1;
					child = (CHierarchyBase*)node2;
					parent->addChild(child);
				}
			}
		}
	}
	if (result) result = reader->pop("Hierarchy");

	if (result) result = reader->push("Connections");
	if (result)
	{
		int numConnections;
		result = reader->read("Num", numConnections);
		CConnection connection;
		for (i=0; i<numConnections && result; i++)
		{
			if (result = connection.read(reader))
			{
				CBase* srcNode = getNode(connection.srcNodeName);
				CBase* dstNode = getNode(connection.dstNodeName);
				if (srcNode != NULL && dstNode != NULL)
				{
					result = dstNode->addConnection(srcNode, 
																					connection.srcAttributeName, 
																					connection.dstAttributeName);
				}
			}
		}
	}
	if (result) result = reader->pop("Connections");

	if (result) result = reader->pop("CGraph");
	return result;
}

bool CGraph::write(IWriter* writer)
{
	if (writer == NULL) return false;
	int i;
	bool result = CHierarchyBase::write(writer);
	if (result) result = writer->push("CGraph");

	if (result) result = writer->push("Nodes");
	if (result)
	{
		stringPCBaseMapIterator it = nodes.begin();
		CBase* node = NULL;
		result = writer->write("Num", (int)nodes.size());
		for (; it != nodes.end() && result; it++)
		{
			if (result = writer->push("Node"))
			{
				node = it->second;
				result = writer->write("Type", node->getTypeName());
				result = node->write(writer);
			}
			if (result) result = writer->pop("Node");
		}
	}
	if (result) result = writer->pop("Nodes");

	if (result) result = writer->push("Hierarchy");
	if (result)
	{
		pCHierarchyBaseVector hierarchy;
		getHierarchy(this, hierarchy);
		result = writer->write("Num", (int)hierarchy.size());
		if (result) result = writeHierarchy(writer, this);
	}
	if (result) result = writer->pop("Hierarchy");

	if (result) result = writer->push("Connections");
	if (result)
	{
		pCConnectionVector connections;
		getConnections(connections);
		result = writer->write("Num", (int)connections.size());
		for (i=0; i<connections.size() && result; i++)
		{
			result = connections[i]->write(writer);
		}
	}
	if (result) result = writer->pop("Connections");

	if (result) result = writer->pop("CGraph");
	return result;
}

void CGraph::update(float time)
{
	if (dirty)
	{
		connectAll();
		dirty = false;
	}
	// As the time goes by... store it!
	lastTime = currentTime;
	float delta = time - lastTime;
	currentTime += delta;
	// Update the nodes... if the graph is not paused!
	if (!pause)
	{
		stringPCBaseMapIterator it = nodes.begin();
		for (; it != nodes.end(); it++)
		{
			CBase* node = it->second;
			node->update(currentTime);
		}
	}
}

CGraph::CGraph(): pause(false)
{
	typeName = "CGraph";
}

CGraph::~CGraph()
{
	end();
}

bool CGraph::addNode(CBase* node)
{
	bool result = false;
	if (node != NULL)
	{
		result = nodes.insert(stringPCBaseMap::value_type(node->getName(), node)).second;
	}
	dirty = true;
	return result;
}

CBase* CGraph::getNode(const string& name)
{
	CBase* node = NULL;
	if (name == this->name)
	{
		node = this;
		return node;
	}
	stringPCBaseMapIterator it = nodes.find(name);
	if (it != nodes.end())
	{
		node = it->second;
	}
	return node;
}

void CGraph::end()
{
	stringPCBaseMapIterator it = nodes.begin();
	for (; it != nodes.end(); it++)
	{
		CBase* node = it->second;
		delete node;
	}
	nodes.clear();
	dirty = true;
}

bool CGraph::connectAll()
{
	bool result = true;
	stringPCBaseMapIterator it = nodes.begin();
	for (; it != nodes.end(); it++)
	{
		result &= it->second->connect(*this);
	}
	return result;
}

bool CGraph::writeHierarchy(IWriter* writer, CHierarchyBase* node)
{
	bool result = true;
	if (writer == NULL || node == NULL) return false;
	CHierarchyBase* parent = node;
	CHierarchyBase* child = NULL;
	for (int i=0; i<parent->getNumChildren() && result; i++)
	{
		child = parent->getChild(i);
		if (parent != NULL && child != NULL)
		{
			result = writer->write("Parent", parent->getName().c_str());
			if (result) result = writer->write("Child", child->getName().c_str());
			if (result) result = writeHierarchy(writer, child);
		}
	}
	return result;
}

void CGraph::getHierarchy(CHierarchyBase* parent, pCHierarchyBaseVector& hierarchy)
{
	if (parent == NULL) return;
	for (int i=0; i< parent->getNumChildren(); i++)
	{
		CHierarchyBase* child = parent->getChild(i);
		getHierarchy(child, hierarchy);
	}
	if (parent->getParent() != NULL)
	{
		hierarchy.push_back(parent);
	}
}

void CGraph::getConnections(pCConnectionVector& connections)
{
	stringPCBaseMapIterator it = nodes.begin();
	for (; it != nodes.end(); it++)
	{
		CBase* node = it->second;
		for (int i=0; i<node->getNumConnections(); i++)
		{
			CConnection* connection = node->getConnection(i);
			if (connection != NULL)
			{
				connections.push_back(connection);
			}
		}
	}
}
