/*  Eside 3D library, version 5.
    Copyright (C) 2002 Iker Jamardo Zugaza <ijamardo@eside.deusto.es>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    This library has been slightly modified by Grzegorz Adam Hankiewicz
    <gradha@terra.es> and is therefor not exactly the original, which you
    should be able to get from Iker by email.
*/

#include "E3D_CHierarchyBase.h"
#include "E3D_CRender.h"
#include "E3D_IReader.h"
#include "E3D_IWriter.h"

bool CHierarchyBase::read(IReader* reader)
{
	if (reader == NULL) return false;
	bool result = CBase::read(reader);
	if (result) result = reader->push("CHierarchyBase");
	if (result) result = reader->read("visible", visible);
	if (result) result = reader->pop("CHierarchyBase");
	return result;
}

bool CHierarchyBase::write(IWriter* writer)
{
	if (writer == NULL) return false;
	bool result = CBase::write(writer);
	if (result) result = writer->push("CHierarchyBase");
	if (result) result = writer->write("visible", visible);
	if (result) result = writer->pop("CHierarchyBase");
	return result;
}

void CHierarchyBase::update(float time)
{
	lastTime = currentTime;
	currentTime = time;
}

float* CHierarchyBase::getAttribute(const string& attributeName)
{
	float* attrib = NULL;
	if ((attrib = CBase::getAttribute(attributeName)) == NULL)
	{
		if (attributeName == "visibility")
		{
			attrib = &visible;
		}
	}
	return attrib;
}

CHierarchyBase::CHierarchyBase(): renderBoundingBox(false),															
																	visible(true),
																	parent(NULL),
																	numInstances(0)
{
	typeName = "CHierarchyBase";
	worldMatrix.identity();
}

bool CHierarchyBase::updateHierarchy(CMatrix4D& parentWorldMatrix, CBoundingBox& parentBoundingBox, bool parentDirty)
{
	bool updated = dirty | parentDirty;
	// If something changed, our world marix is the same as our father's
	if (updated)
	{
		worldMatrix = parentWorldMatrix;
		worldMatrix.toArray(worldMatrixArray);
	}

		// All the children bounding boxes will be added to this one, so it has to be the 
	// smallest (a bounding box affects only if its bigger than another)
	CBoundingBox concatenatedBoundingBox;
	concatenatedBoundingBox.min = CVector3D(1e9f, 1e9f, 1e9f);
	concatenatedBoundingBox.max = CVector3D(-1e9f, -1e9f, -1e9f);

	// Call to all our children updateHierarchy methods. They are going to go 
	// adding their bounding boxes to my temporal bounding box
	pCHierarchyBaseVectorIterator it = children.begin();
	for(; it != children.end(); it++)
	{
		CHierarchyBase* child = *it;
		updated |= child->updateHierarchy(worldMatrix, concatenatedBoundingBox, updated);
	}

	// If we have changed, or our parent or our children, update the bounding box
	if (updated)
	{
		if (children.size() != 0)
		{
			// Our bounding box is the result of concatenating all my children bounding boxes
			boundingBox = concatenatedBoundingBox;
		}
		else
		{
			// Out bounding box is just my position
			boundingBox.min = boundingBox.max = CVector3D(worldMatrix.w);
		}

		// Update the parentBounding box with ours
		parentBoundingBox += boundingBox;
	}

	// We are not dirty any more
	dirty = false;

	return updated;
}

void CHierarchyBase::render(CRender* render)
{
	if (visible <= 0.5 || render == NULL) return;
	if (renderBoundingBox)
	{
		render->renderBoundingBox(boundingBox);
	}
	pCHierarchyBaseVectorIterator it = children.begin();
	for(; it != children.end(); it++)
	{
		CHierarchyBase* child = *it;
		child->render(render);
	}
}

void CHierarchyBase::setWorldMatrix(const CMatrix4D& worldMatrix)
{
	this->worldMatrix = worldMatrix;
}

void CHierarchyBase::setVisible(float visible)
{
	this->visible = visible;
}

bool CHierarchyBase::addChild(CHierarchyBase* child)
{
	bool result = false;
	if (child != NULL)
	{
		children.push_back(child);
		child->setParent(this);
		result = true;
	}
	return result;
}

void CHierarchyBase::setParent(CHierarchyBase* parent)
{
	this->parent = parent;
}

CMatrix4D CHierarchyBase::getWorldMatrix()
{
	return worldMatrix;
}

float CHierarchyBase::getVisible()
{
	return visible;
}

int CHierarchyBase::getNumChildren()
{
	return children.size();
}

CHierarchyBase* CHierarchyBase::getChild(int i)
{
	CHierarchyBase* child = NULL;
	if (i >=0 && i < children.size())
	{
		child = children[i];
		child->setParent(this);
	}
	return child;
}

CHierarchyBase* CHierarchyBase::getParent()
{
	return parent;
}

CBoundingBox CHierarchyBase::getBoundingBox()
{
	return boundingBox;
}

void CHierarchyBase::setRenderBoundingBox(bool renderBoundingBox)
{
	this->renderBoundingBox = renderBoundingBox;
}

bool CHierarchyBase::getRenderBoundigBox()
{
	return renderBoundingBox;
}

int CHierarchyBase::getNumInstances()
{
	return numInstances;
}

void CHierarchyBase::incNumInstances()
{
	numInstances++;
}

void CHierarchyBase::decNumInstances()
{
	if (numInstances > 0)
	{
		numInstances--;
	}
}
