/*  Eside 3D library, version 5.
    Copyright (C) 2002 Iker Jamardo Zugaza <ijamardo@eside.deusto.es>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    This library has been slightly modified by Grzegorz Adam Hankiewicz
    <gradha@terra.es> and is therefor not exactly the original, which you
    should be able to get from Iker by email.
*/

#include "E3D_CMesh.h"
#include "E3D_IReader.h"
#include "E3D_IWriter.h"
#include "E3D_CGraph.h"
#include "E3D_CShader.h"

#include <sstream>
using std::stringstream;

DEFINE_CREATOR_OBJECT(CMesh)

CPolygon::CPolygon()
{
	typeName = "CPolygon";
}

bool CPolygon::read(IReader* reader)
{
	if (reader == NULL) return false;
	bool result = true;
	if (result) 
	{
		int numVertexIndices;
		result = reader->read("Indices", numVertexIndices);
		clear();
		positionIndices.resize(numVertexIndices);
		normalIndices.resize(numVertexIndices);
		uvCoordIndices.resize(numVertexIndices);
		for (int i=0; result && i<numVertexIndices; i++)
		{
			result = reader->read("PositionIndex", positionIndices[i]);
			result = reader->read("NormalIndex", normalIndices[i]);
			result = reader->read("uvCoordIndex", uvCoordIndices[i]);
		}
	}
	if (result) result = reader->read("Normal", normal);
	return result;
}

bool CPolygon::write(IWriter* writer)
{
	if (writer == NULL) return false;
	bool result = true;
	if (result) 
	{
		result = writer->write("Indices", (int)positionIndices.size());
		for (int i=0; result && i<positionIndices.size(); i++)
		{
			result = writer->write("PositionIndex", positionIndices[i]);
			result = writer->write("NormalIndex", normalIndices[i]);
			result = writer->write("uvCoordIndex", uvCoordIndices[i]);
		}
	}
	if (result) result = writer->write("Normal", normal);
	return result;
}

void CPolygon::setNormal(const CVector3D& normal)
{
	this->normal = normal;
}

void CPolygon::addPositionIndex(int i)
{
	positionIndices.push_back(i);
}

void CPolygon::addNormalIndex(int i)
{
	normalIndices.push_back(i);
}

void CPolygon::addUVCoordIndex(int i)
{
	uvCoordIndices.push_back(i);
}

int CPolygon::getNumIndices()
{
	return positionIndices.size();
}

int CPolygon::getPositionIndex(int i)
{
	int result = -1;
	if (i >= 0 && i < positionIndices.size())
	{
		result = positionIndices[i];
	}
	return result;
}

int CPolygon::getNormalIndex(int i)
{
	int result = -1;
	if (i >= 0 && i < normalIndices.size())
	{
		result = normalIndices[i];
	}
	return result;
}

int CPolygon::getUVCoordIndex(int i)
{
	int result = -1;
	if (i >= 0 && i < uvCoordIndices.size())
	{
		result = uvCoordIndices[i];
	}
	return result;
}

CVector3D CPolygon::getNormal()
{
	return normal;
}

void CPolygon::clear()
{
	positionIndices.clear();
	normalIndices.clear();
	uvCoordIndices.clear();
}

CIndex::CIndex(): start(0), end(0)
{
	typeName = "CIndex";
}

bool CMesh::read(IReader* reader)
{
	if (reader == NULL) return false;
	int i;
	bool result = CHierarchyBase::read(reader);
	if (result) result = reader->push("CMesh");
	if (result)
	{
		int numPositions;
		result = reader->read("Positions", numPositions);
		if (result) 
		{
			positions.clear();
			positions.resize(numPositions);
		}
		for (i=0; result && i<numPositions; i++)
		{
			result = reader->read("Position", positions[i]);
		}
	}
	if (result)
	{
		int numNormals;
		result = reader->read("Normals", numNormals);
		if (result) normals.resize(numNormals);
		for (i=0; result && i<numNormals; i++)
		{
			result = reader->read("Normal", normals[i]);
		}
	}
	if (result)
	{
		int numUVCoords;
		result = reader->read("UVCoords", numUVCoords);
		if (result) uvCoords.resize(numUVCoords);
		for (i=0; result && i<numUVCoords; i++)
		{
			result = reader->read("UVCoord", uvCoords[i]);
		}
	}
	if (result)
	{
		int numColors;
		result = reader->read("Colors", numColors);
		if (result) colors.resize(numColors);
		for (i=0; result && i<numColors; i++)
		{
			result = reader->read("Color", colors[i]);
		}
	}
	if (result)
	{
		int numPolygons;
		result = reader->read("Polygons", numPolygons);
		if (result)
		{
			polygons.clear();
			polygons.resize(numPolygons);
		}
		for (i=0; result && i<numPolygons; i++)
		{
			result = polygons[i].read(reader);
		}
	}
	if (result)
	{
		result = reader->read("Shader", shaderName);
	}
	if (result) result = reader->pop("CMesh");
	return result;
}

bool CMesh::write(IWriter* writer)
{
	if (writer == NULL) return false;
	int i;
	bool result = CHierarchyBase::write(writer);
	if (result) result = writer->push("CMesh");
	if (result)
	{
		result = writer->write("Positions", (int)positions.size());
		for (i=0; result && i<positions.size(); i++)
		{
			result = writer->write("Position", positions[i]);
		}
	}
	if (result)
	{
		result = writer->write("Normals", (int)normals.size());
		for (i=0; result && i<normals.size(); i++)
		{
			result = writer->write("Normal", normals[i]);
		}
	}
	if (result)
	{
		result = writer->write("UVCoords", (int)uvCoords.size());
		for (i=0; result && i<uvCoords.size(); i++)
		{
			result = writer->write("UVCoord", uvCoords[i]);
		}
	}
	if (result)
	{
		result = writer->write("Colors", (int)colors.size());
		for (i=0; result && i<colors.size(); i++)
		{
			result = writer->write("Color", colors[i]);
		}
	}
	if (result)
	{
		result = writer->write("Polygons", (int)polygons.size());
		for (i=0; result && i< polygons.size(); i++)
		{
			result = polygons[i].write(writer);
		}
	}
	if (result)
	{
		result = writer->write("Shader", (shader!=NULL)?shader->getName():"NULL");
	}
	if (result) result = writer->pop("CMesh");
	return result;
}

void CMesh::update(float time)
{
	dirty |= updateConnections(time);

	lastTime = currentTime;
	currentTime = time;
	if (dirty)
	{
		buildMesh();
		computeBoundingBox();
	}
}

bool CMesh::connect(CGraph& graph)
{
	if (shaderName == "NULL" || shaderName == "") return true;
	bool result = false;
	CBase* node = graph.getNode(shaderName);
	if (node != NULL)
	{
		try
		{
			shader = dynamic_cast<CShader*>(node);
			result = true;
		}
		catch(bad_cast bce)
		{
		}
	}
	return result;
}

float* CMesh::getAttribute(const string& attributeName)
{
	float* attrib = NULL;
	if ((attrib = CHierarchyBase::getAttribute(attributeName)) == NULL)
	{
		// Parse the attribute to get the position index and vector attribute.
		// Attribute names should come in this form...
		//		pnts[].pntx
		//		pnts[].pnty
		//		pnts[].pntz
		int start, end, vertexIndex;
		string vertexIndexString, vectorDirString;

		// First, parse the point index
		if ((start	= attributeName.find_first_of("[")) != string::npos &&
				(end		= attributeName.find_first_of("]")) != string::npos)
		{
			vertexIndexString = attributeName.substr(start + 1, end - start - 1);

			// Create a conversor to pass from a string to a int
			stringstream conversor(vertexIndexString);
			conversor >> vertexIndex;

			// Now parse the vector dir
			if ((start	= attributeName.find_first_of(".")) != string::npos &&
					(end		= attributeName.length()) != string::npos)
			{
				vectorDirString = attributeName.substr(start + 1, end - start);
				if (vectorDirString == "pntx")
				{
					attrib = &(positions[vertexIndex].x);
				}
				else if (vectorDirString == "pnty")
				{
					attrib = &(positions[vertexIndex].y);
				}
				else if (vectorDirString == "pntz")
				{
					attrib = &(positions[vertexIndex].z);
				}
			}
		}
	}
	return attrib;
}

bool CMesh::updateHierarchy(CMatrix4D& parentWorldMatrix, CBoundingBox& parentBoundingBox, bool parentDirty)
{
	bool updated = dirty | parentDirty;
	// We allready know our bounding box, but we have to transform it acording
	// to our parent transformations. We shouldn't have children so we just forget
	// about them. See maya library for any help...
	CBoundingBox transformedBoundingBox;
	if (updated)
	{
		// We are do not have any local transformations so we take our parents
		worldMatrix = parentWorldMatrix;
		worldMatrix.toArray(worldMatrixArray);

		// Update our bounding box with our parent transformations
		transformedBoundingBox = worldMatrix * boundingBox;
	}

	parentBoundingBox += transformedBoundingBox;

	dirty = false;

	return updated;
}

void CMesh::render(CRender* render)
{
	if (render == NULL) return;
	if (renderBoundingBox)
	{
		if (parent != NULL)
		{
			render->renderBoundingBox(parent->getBoundingBox());
		}
		else
		{
			render->renderBoundingBox(boundingBox);
		}
	}
	render->setModelMode(mode);
	render->setModelNumIndices(numIndices);
	render->setModelIndices(indexArray);
	render->setModelNumVertices(numVertices);
	render->setModelPositions(positionArray);
	render->setModelNormals(normalArray);
	render->setModelUVCoords(uvCoordArray);
	render->setModelColors(colorArray);
	if (shader != NULL)
	{
		shader->setupToRender(render);
	}
	render->renderModel();
}

CMesh::CMesh(): mode(E3D_POLYGON),
								shader(NULL)
{
	typeName = "CMesh";
	clear();
}

CMesh::~CMesh()
{
	end();
}

void CMesh::addPosition(const CVector3D& position)
{
	positions.push_back(position);
	dirty = true;
}

void CMesh::addNormal(const CVector3D& normal)
{
	normals.push_back(normal);
	dirty = true;
}

void CMesh::addUVCoord(const CUVCoord& uvCoord)
{
	uvCoords.push_back(uvCoord);
	dirty = true;
}

void CMesh::addColor(const CColor& color)
{
	colors.push_back(color);
	dirty = true;
}

void CMesh::addVertex(const CVector3D& position, 
											const CVector3D& normal, 
											const CUVCoord& uvCoord,
											const CColor& color)
{
	positions.push_back(position);
	normals.push_back(normal);
	uvCoords.push_back(uvCoord);
	colors.push_back(color);
	dirty = true;
}

void CMesh::addPolygon(const CPolygon& polygon)
{
	polygons.push_back(polygon);
}

int CMesh::getNumPositions()
{
	return positions.size();
}

CVector3D CMesh::getPosition(int i)
{
	CVector3D result;
	if (i >= 0 && i < positions.size())
	{
		result = positions[i];
	}
	return result;
}

int CMesh::getNumNormals()
{
	return normals.size();
}

CVector3D CMesh::getNormal(int i)
{
	CVector3D result;
	if (i >= 0 && i < normals.size())
	{
		result = normals[i];
	}
	return result;
}

int CMesh::getNumUVCoords()
{
	return uvCoords.size();
}

CUVCoord CMesh::getUVCoord(int i)
{
	CUVCoord result;
	if (i >= 0 && i < uvCoords.size())
	{
		result = uvCoords[i];
	}
	return result;
}

int CMesh::getNumColors()
{
	return colors.size();
}

CColor CMesh::getColor(int i)
{
	CColor result;
	if (i >= 0 && i < colors.size())
	{
		result = colors[i];
	}
	return result;
}

int CMesh::getNumPolygons()
{
	return polygons.size();
}

CPolygon CMesh::getPolygon(int i)
{
	CPolygon result;
	if (i >= 0 && i < polygons.size())
	{
		result = polygons[i];
	}
	return result;
}

void CMesh::clear()
{
	indexArray = NULL;
	positionArray = NULL;
	normalArray = NULL;
	uvCoordArray = NULL;
	colorArray = NULL;
	numVertices = numIndices = 0;
}

void CMesh::end()
{
	delete [] indexArray;
	delete [] positionArray;
	delete [] normalArray;
	delete [] uvCoordArray;
	delete [] colorArray;
	clear();
}

void CMesh::computeBoundingBox()
{
	int size = positions.size();
	// If there are no vertices, then the boundingBox is the position of the mesh
	if (size <= 0)
	{
		boundingBox.min = boundingBox.max = worldMatrix.w;
	}
	else
	{
		// This is the smallest boundingBox ever... ;)
		boundingBox.min = CVector3D(1e9f, 1e9f, 1e9f);
		boundingBox.max = CVector3D(-1e9f, -1e9f, -1e9f);
		// Check every vertex position and get minimun and maximun of them
		for(int i=0; i<size; i++)
		{
			// Bigger than the minimun
			if (boundingBox.min.x > positions[i].x)
			{
				boundingBox.min.x = positions[i].x;
			}
			if (boundingBox.min.y > positions[i].y)
			{
				boundingBox.min.y = positions[i].y;
			}
			if (boundingBox.min.z > positions[i].z)
			{
				boundingBox.min.z = positions[i].z;
			}
			// Smaller than the maximun
			if (boundingBox.max.x < positions[i].x)
			{
				boundingBox.max.x = positions[i].x;
			}
			if (boundingBox.max.y < positions[i].y)
			{
				boundingBox.max.y = positions[i].y;
			}
			if (boundingBox.max.z < positions[i].z)
			{
				boundingBox.max.z = positions[i].z;
			}
		}
	}
}

void CMesh::setShaderName(const string& name)
{
	shaderName = name;
}

string CMesh::getShaderName()
{
	return shaderName;
}

void CMesh::setShader(CShader* shader)
{
	this->shader = shader;
}

CShader* CMesh::getShader()
{
	return shader;
}

bool CMesh::buildMesh()
{
	bool result = false;;
	// Create the correct mesh acording to the requested mode
	switch(mode)
	{
		case E3D_POINTS:
			result = buildMesh_E3D_POINTS();
			break;
		case E3D_LINES:
			result = buildMesh_E3D_LINES();
			break;
		case E3D_TRIANGLES:
		case E3D_POLYGON:
			result = buildMesh_E3D_POLYGON();
			break;
	}
	return result;
}

bool CMesh::buildMesh_E3D_POINTS()
{
	int i;
	end();
	numVertices = positions.size();
	// Create the arrays acording to the vectors and copy for positions
	positionArray = new float[numVertices * 3];
	normalArray = new float[numVertices * 3];
	uvCoordArray = new float[numVertices * 2];
	colorArray = new float[numVertices * 4];
	for (i=0; i<numVertices; i++)
	{
		positionArray[i*3] = positions[i].x; positionArray[i*3+1] = positions[i].y; positionArray[i*3+2] = positions[i].z;
		uvCoordArray[i*2] = uvCoords[i].u; uvCoordArray[i*2+1] = uvCoords[i].v;
		colorArray[i*4] = colors[i].a; colorArray[i*4+1] = colors[i].r; colorArray[i*4+2] = colors[i].g; colorArray[i*4+3] = colors[i].b;
		normalArray[i*3] = normals[i].x; normalArray[i*3+1] = normals[i].y; normalArray[i*3+2] = normals[i].z;
	}
	// Create the array for indices
	// There is just one index and its size is the numver of vertices
	numIndices = 1;
	indexArray = new CIndex[numIndices];
	indexArray[0].start = 0;
	indexArray[0].end = numVertices;

	return true;
}

bool CMesh::buildMesh_E3D_LINES()
{
	bool result = true;
	int i, j;
	end();

	// As we cant know how many vertices and indices are inside all the polygons, use 
	// vectors and at the end copy them to arrays
	CVector3DVector finalPositions;
	CVector3DVector finalNormals;
	CUVCoordVector finalUVCoords;
	CColorVector finalColors;
	CIndexVector finalIndices;

	// starts variable is used to know the beginning of each index
	CIndex index;
	for (i=0; i<polygons.size(); i++)
	{
		for (j=0; j<polygons[i].getNumIndices()-1; j++)
		{
			finalPositions.push_back(positions[polygons[i].getPositionIndex(j)]);
			finalColors.push_back(colors[polygons[i].getPositionIndex(j)]);
			finalNormals.push_back(normals[polygons[i].getNormalIndex(j)]);
			finalUVCoords.push_back(uvCoords[polygons[i].getUVCoordIndex(j)]);
			index.start = numVertices;
			index.end = 2;
			finalIndices.push_back(index);
			numVertices++;
		}
		finalPositions.push_back(positions[polygons[i].getPositionIndex(0)]);
		finalColors.push_back(colors[polygons[i].getPositionIndex(0)]);
		finalNormals.push_back(normals[polygons[i].getNormalIndex(0)]);
		finalUVCoords.push_back(uvCoords[polygons[i].getUVCoordIndex(0)]);
		numVertices++;
	}
	// Copy the vertices now that we know their number
	positionArray = new float[finalPositions.size() * 3];
	uvCoordArray = new float[finalUVCoords.size() * 2];
	colorArray = new float[finalColors.size() * 4];
	normalArray = new float[finalNormals.size() * 3];
	for (i=0; i<finalPositions.size(); i++)
	{
		positionArray[i*3] = finalPositions[i].x; positionArray[i*3+1] = finalPositions[i].y; positionArray[i*3+2] = finalPositions[i].z;
		uvCoordArray[i*2] = finalUVCoords[i].u; uvCoordArray[i*2+1] = finalUVCoords[i].v;
		colorArray[i*4] = finalColors[i].a; colorArray[i*4+1] = finalColors[i].r; colorArray[i*4+2] = finalColors[i].g; colorArray[i*4+3] = finalColors[i].b;
		normalArray[i*3] = finalNormals[i].x; normalArray[i*3+1] = finalNormals[i].y; normalArray[i*3+2] = finalNormals[i].z;
	}
	numIndices = finalIndices.size();
	indexArray = new CIndex[numIndices];
	for (i=0; i<numIndices; i++)
	{
		indexArray[i] = finalIndices[i];
	}
	return result;
}

bool CMesh::buildMesh_E3D_POLYGON()
{
	bool result = true;
	int i, j, starts;
	end();

	// As we cant know how many vertices are inside all the polygons, use a vector
	// and at the end copy it to the array
	CVector3DVector finalPositions;
	CVector3DVector finalNormals;
	CUVCoordVector finalUVCoords;
	CColorVector finalColors;

	// There are as many indices as polygons
	numIndices = polygons.size();
	indexArray = new CIndex[numIndices];

	// starts variable is used to know the beginning of each index
	for (starts = 0, i=0; i<polygons.size(); i++)
	{
		for (j=0; j<polygons[i].getNumIndices(); j++)
		{
			int vertexIndex = polygons[i].getPositionIndex(j);
			finalPositions.push_back(positions[vertexIndex]);
			finalColors.push_back(colors[vertexIndex]);
			int normalIndex = polygons[i].getNormalIndex(j);
			finalNormals.push_back(normals[normalIndex]);
			int uvCoordIndex = polygons[i].getUVCoordIndex(j);
			finalUVCoords.push_back(uvCoords[uvCoordIndex]);
			numVertices++;
		}
		indexArray[i].start = starts;
		starts += indexArray[i].end = polygons[i].getNumIndices();
	}
	// Copy the vertices now that we know their number
	positionArray = new float[finalPositions.size() * 3];
	uvCoordArray = new float[finalUVCoords.size() * 2];
	colorArray = new float[finalColors.size() * 4];
	normalArray = new float[finalNormals.size() * 3];
	for (i=0; i<finalPositions.size(); i++)
	{
		positionArray[i*3] = finalPositions[i].x; positionArray[i*3+1] = finalPositions[i].y; positionArray[i*3+2] = finalPositions[i].z;
		uvCoordArray[i*2] = finalUVCoords[i].u; uvCoordArray[i*2+1] = finalUVCoords[i].v;
		colorArray[i*4] = finalColors[i].a; colorArray[i*4+1] = finalColors[i].r; colorArray[i*4+2] = finalColors[i].g; colorArray[i*4+3] = finalColors[i].b;
		normalArray[i*3] = finalNormals[i].x; normalArray[i*3+1] = finalNormals[i].y; normalArray[i*3+2] = finalNormals[i].z;
	}
	return result;
}

