/*  Eside 3D library, version 5.
    Copyright (C) 2002 Iker Jamardo Zugaza <ijamardo@eside.deusto.es>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    This library has been slightly modified by Grzegorz Adam Hankiewicz
    <gradha@terra.es> and is therefor not exactly the original, which you
    should be able to get from Iker by email.
*/

#include "E3D_CTransform.h"
#include "E3D_CRender.h"
#include "E3D_IReader.h"
#include "E3D_IWriter.h"

DEFINE_CREATOR_OBJECT(CTransform)

bool CTransform::read(IReader* reader)
{
	if (reader == NULL) return false;
	bool result = CHierarchyBase::read(reader);
	if (result) result = reader->push("CTransform");
	if (result) result = reader->read("Translate", translate);
	if (result) result = reader->read("Rotate", rotate);
	if (result) result = reader->read("Scale", scale);
	if (result) result = reader->read("Pivot", pivot);
	if (result) result = reader->pop("CTransform");
	return result;
}

bool CTransform::write(IWriter* writer)
{
	if (writer == NULL) return false;
	bool result = CHierarchyBase::write(writer);
	if (result) result = writer->push("CTransform");
	if (result) result = writer->write("Translate", translate);
	if (result) result = writer->write("Rotate", rotate);
	if (result) result = writer->write("Scale", scale);
	if (result) result = writer->write("Pivot", pivot);
	if (result) result = writer->pop("CTransform");
	return result;
}

void CTransform::update(float time)
{
	dirty |= updateConnections(time);

	lastTime = currentTime;
	currentTime = time;

	if (dirty)
	{
		// Calculate the local matrix acording to actual translate, rotate and scale
		if (pivot == CVector3D(0, 0, 0))
		{
			localMatrix.translateRotateScale(translate, rotate, scale);
		}
		else
		{
			// It is important to locate the transfrm acording to the pivot position. To do
			// so, the final translation comes by the position plus pivot offset.
			// Then the rotations and scalation operates over the negative pivot position.
			// It is just like a parent child matrix product
			CMatrix4D pivotMatrix;
			pivotMatrix.identity();
			pivotMatrix.w = CVector4D(-pivot.x, -pivot.y, -pivot.z, 1);
			localMatrix.translateRotateScale(translate + pivot, rotate, scale);
			localMatrix = localMatrix * pivotMatrix;
		}
	}
}

float* CTransform::getAttribute(const string& attributeName)
{
	float* attrib = NULL;
	if ((attrib = CHierarchyBase::getAttribute(attributeName)) == NULL)
	{
		if (attributeName == "translateX")
		{
			attrib = &(translate.x);
		}
		else if (attributeName == "translateY")
		{
			attrib = &(translate.y);
		}
		else if (attributeName == "translateZ")
		{
			attrib = &(translate.z);
		}
		else if (attributeName == "rotateX")
		{
			attrib = &(rotate.x);
		}
		else if (attributeName == "rotateY")
		{
			attrib = &(rotate.y);
		}
		else if (attributeName == "rotateZ")
		{
			attrib = &(rotate.z);
		}
		else if (attributeName == "scaleX")
		{
			attrib = &(scale.x);
		}
		else if (attributeName == "scaleY")
		{
			attrib = &(scale.y);
		}
		else if (attributeName == "scaleZ")
		{
			attrib = &(scale.z);
		}
	}
	return attrib;
}

void CTransform::render(CRender* render)
{
	if (!visible || render == NULL) return;
	render->pushModelMatrix(worldMatrixArray);
	CHierarchyBase::render(render);
	render->popModelMatrix();
}

bool CTransform::updateHierarchy(CMatrix4D& parentWorldMatrix, CBoundingBox& parentBoundingBox, bool parentDirty)
{
	bool updated = dirty | parentDirty;

	if (updated)
	{
		// Just to avoid a multiplication with identity... jeje
		if (parent == NULL)
		{
			// We are in the top level, so our world matrix and local matrix are the same
			worldMatrix = localMatrix;
		}
		else
		{
			// We are inside a hierarchy so our reference is our parent and its world
			// transformations concatenated to our local transformations give us our world transformations
			// Figurese que movida!
			worldMatrix = parentWorldMatrix * localMatrix;
		}
		worldMatrix.toArray(worldMatrixArray);
	}

	// All the children bounding boxes will be added to this one, so it has to be the 
	// smallest (a bounding box affects only if its bigger than another)
	CBoundingBox concatenatedBoundingBox;
	concatenatedBoundingBox.min = CVector3D(1e9f, 1e9f, 1e9f);
	concatenatedBoundingBox.max = CVector3D(-1e9f, -1e9f, -1e9f);

	// Call to all our children updateHierarchy methods. They are going to go 
	// adding their bounding boxes to my temporal bounding box
	pCHierarchyBaseVectorIterator it = children.begin();
	for(; it != children.end(); it++)
	{
		CHierarchyBase* child = *it;
		updated |= child->updateHierarchy(worldMatrix, concatenatedBoundingBox, updated);
	}

	// If we have changed, or our parent or our children has, update the bounding box
	if (updated)
	{
		if (children.size() != 0)
		{
			// Our bounding box is the result of concatenating all my children bounding boxes
			boundingBox = concatenatedBoundingBox;
		}
		else
		{
			// Out bounding box is just my position
			boundingBox.min = boundingBox.max = worldMatrix.w;
		}

		// Update the parentBounding box with ours
		parentBoundingBox += boundingBox;
	}

	// We are no dirty anymore
	dirty = false;

	return updated;
}

CTransform::CTransform()
{
	typeName = "CTransform";
	scale = CVector3D(1, 1, 1);
	localMatrix.identity();
}

void CTransform::setTranslate(const CVector3D& translate)
{
	dirty = true;
	this->translate = translate;
}

void CTransform::setRotate(const CVector3D& rotate)
{
	dirty = true;
	this->rotate = rotate;
}

void CTransform::setScale(const CVector3D& scale)
{
	dirty = true;
	this->scale = scale;
}

void CTransform::setPivot(const CVector3D& pivot)
{
	this->pivot = pivot;
}

CVector3D CTransform::getTranslate()
{
	return translate;
}

CVector3D CTransform::getRotate()
{
	return rotate;
}

CVector3D CTransform::getScale()
{
	return scale;
}

CVector3D CTransform::getPivot()
{
	return pivot;
}
