/*  Eside 3D library, version 5.
    Copyright (C) 2002 Iker Jamardo Zugaza <ijamardo@eside.deusto.es>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    This library has been slightly modified by Grzegorz Adam Hankiewicz
    <gradha@terra.es> and is therefor not exactly the original, which you
    should be able to get from Iker by email.
*/

#ifndef _E3D_CBASE_H_
#define _E3D_CBASE_H_

#pragma warning (disable: 4786 4503)

#include "E3D_CObject.h"
#include "E3D_CConnection.h"

#include <vector>
using std::vector;
#include <map>
using std::map;

using std::bad_cast;

class IReader;
class IWriter;
class CGraph;

// ============================================================================
// The base class for any node. Stores some useful data and the main virtual
// method to update any node.
// ============================================================================
class CBase: public CObject
{
public:
// From CObject
	// ============================================================================
	// An operator to cast any object to a string. Polimorph it to give it a new
	// behaviour that fits your needs.
	// ============================================================================
	virtual operator string() const;

	// ============================================================================
	// An operator to compare the equality of two objects. Polimorph it to give
	// a new behaviour that fits your needs.
	// ============================================================================
	virtual bool operator==(const CObject& obj) const;

// Own
	// ============================================================================
	// Constructor. Initializes all the attributes.
	// ============================================================================
	CBase();

	// ============================================================================
	// Destructor. Needed to be virtual in order to destroy inherited objects.
	// ============================================================================
	virtual ~CBase();

	// ============================================================================
	// Method to read all the data related with this class from a reader interface.
	// ============================================================================
	virtual bool read(IReader* reader);

	// ============================================================================
	// Method to write all the data related with this class to a writer interface
	// ============================================================================
	virtual bool write(IWriter* writer);

	// ============================================================================
	// Updates any object. Polimorph it to give it your own behaivour.
	// In this case just updates the timing.
	// ============================================================================
	virtual void update(float time);

	// ============================================================================
	// This method should be implemented in all the classes that wants to be connected.
	// Returns the pointer to the attribute specified by its name.
	// ============================================================================
	virtual float* getAttribute(const string& attributeName);

	// ============================================================================
	// Use this method to connect some nodes that cannot find their reference.
	// Polimorph it and use it only where the class needs it.
	// ============================================================================
	virtual bool connect(CGraph& graph);

	// ============================================================================
	// Give the object a name.
	// ============================================================================
	void setName(const string& name);

	// ============================================================================
	// Set the start time.
	// ============================================================================
	void setStartTime(float startTime);

	// ============================================================================
	// Set the end time.
	// ============================================================================
	void setEndTime(float endTime);

	// ============================================================================
	// Get the object name.
	// ============================================================================
	string getName();

	// ============================================================================
	// Get the starting time.
	// ============================================================================
	float getStartTime();

	// ============================================================================
	// Get the ending time.
	// ============================================================================
	float getEndTime();

	// ============================================================================
	// ============================================================================
	bool addConnection(CBase* srcNode, const string& srcAttribute, const string& dstAttribute);

	// ============================================================================
	// ============================================================================
	void removeConnections();

	// ============================================================================
	// ============================================================================
	int getNumConnections();

	// ============================================================================
	// ============================================================================
	CConnection* getConnection(int index);

protected:
	bool updateConnections(float time);

	bool dirty;
	float startTime, endTime, currentTime, lastTime;
	string name;
	pCConnectionVector connections;
};

typedef vector<CBase*> pCBaseVector;
typedef pCBaseVector::iterator pCBaseVectorIterator;

typedef map<string, CBase*> stringPCBaseMap;
typedef stringPCBaseMap::iterator stringPCBaseMapIterator;

#endif