/*  Eside 3D library, version 5.
    Copyright (C) 2002 Iker Jamardo Zugaza <ijamardo@eside.deusto.es>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    This library has been slightly modified by Grzegorz Adam Hankiewicz
    <gradha@terra.es> and is therefor not exactly the original, which you
    should be able to get from Iker by email.
*/

#ifndef _E3D_CHIERARCHYBASE_H_
#define _E3D_CHIERARCHYBASE_H_

#include "E3D_CBase.h"
#include "E3D_CHierarchyBase.h"
#include "E3D_CMatrix4D.h"
#include "E3D_CBoundingBox.h"

class CHierarchyBase;
class IReader;
class IWriter;
class CRender;

typedef vector<CHierarchyBase*> pCHierarchyBaseVector;
typedef pCHierarchyBaseVector::iterator pCHierarchyBaseVectorIterator;


// ============================================================================
// The base class for any node inside the scene hierarchy. This means that
// any node inherited from the CHierarchyBase is going to be in the 3D space
// and it will have a WorldMatrix. The main virtual methods are updateHierarchy
// and render. To implement the graph, any hierarchycal node has a parent and 
// a container of children.
// ============================================================================
class CHierarchyBase: public CBase
{
public:
// From CBase
	// ============================================================================
	// Method to read all the data related with this class from a reader interface.
	// ============================================================================
	virtual bool read(IReader* reader);

	// ============================================================================
	// Method to write all the data related with this class to a writer interface
	// ============================================================================
	virtual bool write(IWriter* writer);

	// ============================================================================
	// Updates any object. Polimorph it to give it your own behaivour.
	// In this case updates all the children objects.
	// ============================================================================
	virtual void update(float time);

	// ============================================================================
	// This method should be implemented in all the classes that wants to be connected.
	// Returns the pointer to the attribute specified by its name.
	// ============================================================================
	virtual float* getAttribute(const string& attributeName);

// Own
	// ============================================================================
	// The constructor to initialize some internal attributes.
	// ============================================================================
	CHierarchyBase();

	// ============================================================================
	// Updates the scene hirarchy updating all the children. The local world matrix
	// is passed to all the children and a flag to indicate either the parent
	// has change or not.
	// Polimorph it to give it a new behaivour.
	// ============================================================================
	virtual bool updateHierarchy(CMatrix4D& parentWorldMatrix, CBoundingBox& parentBoundingBox, bool parentDirty);

	// ============================================================================
	// Render hierarchically. 
	// Polimorph it to give it a new behaivour.
	// ============================================================================
	virtual void render(CRender* render);

	// ============================================================================
	// Give a new world matrix from outside.
	// ============================================================================
	void setWorldMatrix(const CMatrix4D& worldMatrix);

	// ============================================================================
	// Indicates if the node is visible or not. This means that neither this node
	// or its children will be rendered.
	// ============================================================================
	void setVisible(float visible);

	// ============================================================================
	// Add a child to the children vector. This is going to be the parent of the
	// new child.
	// ============================================================================
	bool addChild(CHierarchyBase* child);

	// ============================================================================
	// Set the parent of thid node. This node is added as a child to the parent node.
	// ============================================================================
	void setParent(CHierarchyBase* parent);

	// ============================================================================
	// Get the world matrix.
	// ============================================================================
	CMatrix4D getWorldMatrix();

	// ============================================================================
	// Get the visible flag to know if this node is rendered or not.
	// ============================================================================
	float getVisible();

	// ============================================================================
	// Get the number of children that this node has.
	// ============================================================================
	int getNumChildren();

	// ============================================================================
	// Get a children node by its index. Return NULL if the index was incorrect.
	// ============================================================================
	CHierarchyBase* getChild(int i);

	// ============================================================================
	// Get the parent pointer.
	// ============================================================================
	CHierarchyBase* getParent();

	// ============================================================================
	// ============================================================================
	CBoundingBox getBoundingBox();

	// ============================================================================
	// ============================================================================
	void setRenderBoundingBox(bool renderBoundingBox);

	// ============================================================================
	// ============================================================================
	bool getRenderBoundigBox();

	int getNumInstances();
	void incNumInstances();
	void decNumInstances();

protected:
	bool renderBoundingBox;
	CBoundingBox boundingBox;
	CMatrix4D worldMatrix;
	float worldMatrixArray[16];
	float visible;
	CHierarchyBase* parent;
	pCHierarchyBaseVector children;
	int numInstances;
};


#endif