/*  Eside 3D library, version 5.
    Copyright (C) 2002 Iker Jamardo Zugaza <ijamardo@eside.deusto.es>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    This library has been slightly modified by Grzegorz Adam Hankiewicz
    <gradha@terra.es> and is therefor not exactly the original, which you
    should be able to get from Iker by email.
*/

#ifndef _E3D_CMESH_H_
#define _E3D_CMESH_H_

#include "E3D_CHierarchyBase.h"
#include "E3D_CUVCoord.h"
#include "E3D_CColor.h"
#include "E3D_CRender.h"

typedef vector<int> intVector;
typedef intVector::iterator intVectorIterator;

// ============================================================================
// ============================================================================
class CPolygon: public CObject
{
public:
	// ============================================================================
	// ============================================================================
	CPolygon();

	// ============================================================================
	// Method to read all the data related with this class from a reader interface.
	// ============================================================================
	bool read(IReader* reader);

	// ============================================================================
	// Method to write all the data related with this class to a writer interface
	// ============================================================================
	bool write(IWriter* writer);

	// ============================================================================
	// Set the polygon normal.
	// ============================================================================
	void setNormal(const CVector3D& normal);

	// ============================================================================
	// Ad an index to a vertiex position inside this polygon.
	// ============================================================================
	void addPositionIndex(int i);

	void addNormalIndex(int i);

	void addUVCoordIndex(int i);

	// ============================================================================
	// Get the number of indices that this polygon has (number of vertices)
	// ============================================================================
	int getNumIndices();

	// ============================================================================
	// Get a vertex index. Returns -1 if the parameter is less than 0 or bigger or equal
	// than the number of indices.
	// ============================================================================
	int getPositionIndex(int i);

	// ============================================================================
	// Get a vertex index. Returns -1 if the parameter is less than 0 or bigger or equal
	// than the number of indices.
	// ============================================================================
	int getNormalIndex(int i);

	// ============================================================================
	// Get a vertex index. Returns -1 if the parameter is less than 0 or bigger or equal
	// than the number of indices.
	// ============================================================================
	int getUVCoordIndex(int i);

	// ============================================================================
	// Get the normal of the polygon.
	// ============================================================================
	CVector3D getNormal();

	// ============================================================================
	// ============================================================================
	void clear();

protected:
	CVector3D normal;
	intVector positionIndices;
	intVector normalIndices;
	intVector uvCoordIndices;
};

typedef vector<CPolygon> CPolygonVector;
typedef CPolygonVector::iterator CPolygonVectorIterator;

// ============================================================================
// ============================================================================
class CIndex: public CObject
{
public:
	CIndex();
	int start, end;
};

typedef vector<CIndex> CIndexVector;
typedef CIndexVector::iterator CIndexVectorIterator;

class IReader;
class IWriter;
class CShader;

// ============================================================================
// The class to store all the data related with a 3D geometry.
// Vertices positions, normals, texture coordinates, vertices colors, polygons...
// It is a hierarchy node so it could be affected by a transform matrix.
// ============================================================================
class CMesh: public CHierarchyBase
{
public:
// From CBase
	// ============================================================================
	// Method to read all the data related with this class from a reader interface.
	// ============================================================================
	virtual bool read(IReader* reader);

	// ============================================================================
	// Method to write all the data related with this class to a writer interface
	// ============================================================================
	virtual bool write(IWriter* writer);

	// ============================================================================
	// Updates any object. Polimorph it to give it your own behaivour.
	// In this case creates the corresponding arrays of geometry data if the mesh is
	// dirty in any way.
	// ============================================================================
	virtual void update(float time);

	// ============================================================================
	// This method should be implemented in all the classes that wants to be connected.
	// Returns the pointer to the attribute specified by its name.
	// ============================================================================
	virtual float* getAttribute(const string& attributeName);

	// ============================================================================
	// Use this method to connect some nodes that cannot find their reference.
	// Polimorph it and use it only where the class needs it.
	// ============================================================================
	virtual bool connect(CGraph& graph);

// From CHierarchyBase
	// ============================================================================
	// Updates the scene hirarchy updating all the children. The local world matrix
	// is passed to all the children and a flag to indicate either the parent
	// has change or not.
	// Polimorph it to give it a new behaivour.
	// In this case just takes the parent world matrix as its own world matrix.
	// ============================================================================
	virtual bool updateHierarchy(CMatrix4D& parentWorldMatrix, CBoundingBox& parentBoundingBox, bool parentDirty);

	// ============================================================================
	// Render hierarchically. 
	// Polimorph it to give it a new behaivour.
	// In this case renders all the geometry (and do nothing with the children).
	// ============================================================================
	virtual void render(CRender* render);

// Own
	// ============================================================================
	// ============================================================================
	CMesh();

	// ============================================================================
	// ============================================================================
	~CMesh();

	// ============================================================================
	// Add a vertex 3D position.
	// ============================================================================
	void addPosition(const CVector3D& position);

	// ============================================================================
	// Add a vertex 3D normal.
	// ============================================================================
	void addNormal(const CVector3D& normal);

	// ============================================================================
	// Add a vertex UV coord (for texture).
	// ============================================================================
	void addUVCoord(const CUVCoord& uvCoord);

	// ============================================================================
	// Add a vertex color.
	// ============================================================================
	void addColor(const CColor& color);

	// ============================================================================
	// Add a complete vertex (position, normal, uv coord and color).
	// ============================================================================
	void addVertex(	const CVector3D& position, 
									const CVector3D& normal, 
									const CUVCoord& uvCoord,
									const CColor& color);

	// ============================================================================
	// Add a polygon with its indices pointing to this mesh vertices.
	// ============================================================================
	void addPolygon(const CPolygon& polygon);

	// ============================================================================
	// Indicates how many vertice positions are there inside this geometry.
	// ============================================================================
	int getNumPositions();

	// ============================================================================
	// Get a concrete position. There is an error control check.
	// ============================================================================
	CVector3D getPosition(int i);

	// ============================================================================
	// Indicates how many normals are there inside this geometry (one per vertex?)
	// ============================================================================
	int getNumNormals();

	// ============================================================================
	// Get a concrete normal. There is an error control check.
	// ============================================================================
	CVector3D getNormal(int i);

	// ============================================================================
	// Indicates how many UV coordinates are there inside this geometry. (one per vertex?)
	// ============================================================================
	int getNumUVCoords();

	// ============================================================================
	// Get a concrete UV coordinate. There is an error control check.
	// ============================================================================
	CUVCoord getUVCoord(int i);

	// ============================================================================
	// Indicates how many colors are there inside this geometry. (one per vertex?)
	// ============================================================================
	int getNumColors();

	// ============================================================================
	// Get a concrete color. There is an error control check.
	// ============================================================================
	CColor getColor(int i);

	// ============================================================================
	// Indicates how many faces or polygons are inside this geometry.
	// ============================================================================
	int getNumPolygons();

	// ============================================================================
	// Get a concrete polygon. There is an error control check.
	// ============================================================================
	CPolygon getPolygon(int i);

	// ============================================================================
	// ============================================================================
	void setMode(E3D_MODE mode) { this->mode = mode; }

	// ============================================================================
	// ============================================================================
	void computeBoundingBox();

	// ============================================================================
	// ============================================================================
	void setShaderName(const string& name);

	// ============================================================================
	// ============================================================================
	string getShaderName();

	// ============================================================================
	// ============================================================================
	void setShader(CShader* shader);

	// ============================================================================
	// ============================================================================
	CShader* getShader();

protected:
	void clear();
	void end();
	bool buildMesh();
	bool buildMesh_E3D_POINTS();
	bool buildMesh_E3D_LINES();
	bool buildMesh_E3D_POLYGON();

	// Original mesh data
	CVector3DVector positions;
	CVector3DVector normals;
	CUVCoordVector uvCoords;
	CColorVector colors;
	CPolygonVector polygons;

	// To call to a particular build method
	E3D_MODE mode;

	CShader* shader;
	string shaderName;

	// Render mesh data
	int numVertices;
	int numIndices;
	CIndex* indexArray;
	float* positionArray;
	float* normalArray;
	float* uvCoordArray;
	float* colorArray;
};

DEFINE_CREATOR_CLASS(CMesh)

#endif
