/*  Eside 3D library, version 5.
    Copyright (C) 2002 Iker Jamardo Zugaza <ijamardo@eside.deusto.es>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    This library has been slightly modified by Grzegorz Adam Hankiewicz
    <gradha@terra.es> and is therefor not exactly the original, which you
    should be able to get from Iker by email.
*/

#ifndef _E3D_CTRANSFORM_H_
#define _E3D_CTRANSFORM_H_

#include "E3D_CHierarchyBase.h"
#include "E3D_CMatrix4D.h"

class CRender;
class IReader;
class IWriter;

// ============================================================================
// The class to perform 3D transformations. As it is a 3D object, inherits from
// CHierarchyBase and needs to perform its operations to actualize its localMatrix
// with the parentWorldMatrix passed in the updateHierarchy method.
// ============================================================================
class CTransform: public CHierarchyBase
{
public:

// From CBase
	// ============================================================================
	// Method to read all the data related with this class from a reader interface.
	// ============================================================================
	virtual bool read(IReader* reader);

	// ============================================================================
	// Method to write all the data related with this class to a writer interface
	// ============================================================================
	virtual bool write(IWriter* writer);

	// ============================================================================
	// ============================================================================
	virtual void update(float time);

	// ============================================================================
	// This method should be implemented in all the classes that wants to be connected.
	// Returns the pointer to the attribute specified by its name.
	// ============================================================================
	virtual float* getAttribute(const string& attributeName);

// From CHierarchyBase
	// ============================================================================
	// Updates the scene hirarchy updating all the children. The local world matrix
	// is passed to all the children and a flag to indicate either the parent
	// has change or not.
	// Polimorph it to give it a new behaivour.
	// In this case concatenates the local matrix with the parent world matrix
	// to calculate the correct node world matrix.
	// ============================================================================
	virtual bool updateHierarchy(CMatrix4D& parentWorldMatrix, CBoundingBox& parentBoundingBox, bool parentDirty);

	// ============================================================================
	// Render hierarchically. 
	// Polimorph it to give it a new behaivour.
	// In this case pushes actual matrix to let the children render themselves
	// and the restores the last matrix popping it.
	// ============================================================================
	virtual void render(CRender* render);

// Own
	// ============================================================================
	// Constructor to initilize some data.
	// ============================================================================
	CTransform();

	// ============================================================================
	// Set the translation values.
	// ============================================================================
	void setTranslate(const CVector3D& translate);

	// ============================================================================
	// Set the rotation values.
	// ============================================================================
	void setRotate(const CVector3D& rotate);

	// ============================================================================
	// Set the scalation values.
	// ============================================================================
	void setScale(const CVector3D& scale);

	// ============================================================================
	// Set the point where rotate around and scale from.
	// ============================================================================
	void setPivot(const CVector3D& pivot);

	// ============================================================================
	// Get the current translation values.
	// ============================================================================
	CVector3D getTranslate();

	// ============================================================================
	// Get the current rotation values.
	// ============================================================================
	CVector3D getRotate();

	// ============================================================================
	// Get the current scalation values.
	// ============================================================================
	CVector3D getScale();

	// ============================================================================
	// Get the point where rotate around and scale from.
	// ============================================================================
	CVector3D getPivot();

protected:
	CVector3D pivot;
	CMatrix4D localMatrix;
	CVector3D translate;
	CVector3D rotate;
	CVector3D scale;
};

typedef vector<CTransform*> pCTransformVector;
typedef pCTransformVector::iterator pCTransformVectorIterator;

DEFINE_CREATOR_CLASS(CTransform)

#endif