/*  Katapult, version 1.
    Copyright (C) 2002 Grzegorz Adam Hankiewicz <gradha@terra.es> and
    Jon Lpez de Dicastillo <darwater@hotmail.com>.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "Earth.h"

#include "E3D_CScene.h"

#include <assert.h>

/* Constructor of a piece of earth. It creates the object
 * "Earth", which can be later retrieved and manipulated. At the
 * moment it only creates a plain terrain with the players' zones.
 * Typical code to generate the earth's mesh:

	CEarth *earth = new CEarth(10, 10);
	scene.addNode(earth->mesh);

 */
CEarth::CEarth(CScene &scene, float width, float height)
{
	w = width;
	h = height;

	mesh = new CMesh();

	#define DO_VERTEX(X,Y,Z,R,G,B,U,V) mesh->addVertex(CVector3D(X,Y,Z),\
		CVector3D(0,0,1), CUVCoord(U,V), CColor(R,G,B,1))

	// The total world plane
	DO_VERTEX(0,0,0,								0,0,0,0,0);
	DO_VERTEX(width,0,0,							0,0,0,100,0);
	DO_VERTEX(width,height,0,						0,0,0,100,100);
	DO_VERTEX(0,height,0,							0,0,0,0,100);

	// Red player zone
	float left = width * (1 - ZONE_WIDTH) / 2;
	float right = width - left;
	float bottom = height * ZONE_BOTTOM;
	float top = height * ZONE_TOP;
	DO_VERTEX(left, bottom,0,						1,0,0,0,0);
	DO_VERTEX(right,bottom,0,						1,0,0,0,0);
	DO_VERTEX(right,top,0,							1,0,0,0,0);
	DO_VERTEX(left, top,0,							1,0,0,0,0);

	// Blue player zone
	bottom = height - bottom;
	top = height - top;
	DO_VERTEX(left, top,0,							0,0,1,0,0);
	DO_VERTEX(right,top,0,							0,0,1,0,0);
	DO_VERTEX(right,bottom,0,						0,0,1,0,0);
	DO_VERTEX(left, bottom,0,						0,0,1,0,0);

	CPolygon polygon;
#define DO_QUAD(P1,P2,P3,P4) {\
	polygon.addPositionIndex(P1);\
	polygon.addPositionIndex(P2);\
	polygon.addPositionIndex(P3);\
	polygon.addPositionIndex(P4);\
	polygon.addPositionIndex(P1);\
	mesh->addPolygon(polygon);\
	polygon.clear();}

	DO_QUAD(0,1,2,3);
	DO_QUAD(4,5,6,7);		/*Cuadrados de Zonas*/
	DO_QUAD(8,9,10,11);

	mesh->setMode(E3D_LINES); // Quitar cuando funcione textura
	scene.addNode(mesh);
	scene.getCTransform("Cylinder")->addChild(mesh);
}

/* Destructor of a piece of earth.
 */
CEarth::~CEarth()
{
	delete mesh;
}

float CEarth::width(void)		{ return w; };
float CEarth::height(void)		{ return h; };

/* Returns the corner position of a zone. num shuold be 0 (red)
 * or 1 (blue), and corner can be 0 (the one with lesser 
 * coordinates) or 1 (it's opposite).
 */
CVector3D CEarth::zone(int num, int corner)
{
	assert(num == 0 || num == 1);
	assert(corner == 0 || corner == 1);

	float left = w * (1 - ZONE_WIDTH) / 2;
	float right = w - left;
	float bottom = h * ZONE_BOTTOM;
	float top = h * ZONE_TOP;

	if(num) {
		bottom = h - h * ZONE_TOP;
		top = h - h * ZONE_BOTTOM;
	}

	if(corner)
		return CVector3D(right, top, 0);
	else
		return CVector3D(left, bottom, 0);
}

/* Returns the width of the player's zone */
float CEarth::zoneWidth(void)
{
	return w * ZONE_WIDTH;
}

/* Returns the height of the player's zone */
float CEarth::zoneHeight(void)
{
	return h * (ZONE_TOP - ZONE_BOTTOM);
}
