/*  Katapult, version 1.
    Copyright (C) 2002 Grzegorz Adam Hankiewicz <gradha@terra.es> and
    Jon Lpez de Dicastillo <darwater@hotmail.com>.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "Kamera.h"

#include "E3D_CTransform.h"
#include "E3D_CCamera.h"
#include "E3D_CVector3D.h"
#include "E3D_CScene.h"
#include "E3D_CMath.h"
#include "E3D_CMesh.h"

#include <assert.h>

CKamera::~CKamera()
{
}

/* Creates a camera. Requires a pointer to the scene in order to
 * insert itself in it. You can create only one single camera.
 */
CKamera::CKamera(CScene &scene)
{
	cameraParent = new CTransform();
	cameraChild = new CTransform();
	camera = new CCamera();

	cameraParent->setName("CameraParent");
	cameraChild->setName("CameraChild");
	camera->setName("Camera");

	scene.addNode(cameraParent);
	scene.addNode(cameraChild);
	scene.addNode(camera);

	scene.addChild(cameraParent);
	cameraParent->addChild(cameraChild);
	cameraChild->addChild(camera);
	scene.setActualCamera(camera);
}

/* Places the camera at the specified location in world coordinates
 */
void CKamera::move(float x, float y, float z)
{
	move(CVector3D(x, y, z));
}

/* Places the camera at the specified location in world coordinates
 */
void CKamera::move(const CVector3D &pos)
{
	this->pos = pos;
	updateAngles();
}

/* Makes the camera point at the specified location in world coordinates
 */
void CKamera::point(float x, float y, float z)
{
	point(CVector3D(x, y, z));
}

/* Makes the camera point at the specified location in world coordinates
 */
void CKamera::point(const CVector3D &pos)
{
	cameraParent->setTranslate(pos);
	updateAngles();
}

/* After moving the camera or making it poit somewhere, the 'real'
 * CTransforms have to be updated correctly to show the new position.
 * This function calculates the new rotation angles of the parent
 * CTransform and the distance of the child CTransform.
 */
void CKamera::updateAngles(void)
{
	CVector3D dif = pos - cameraParent->getTranslate();
	double len = sqrt(dif.x*dif.x + dif.y*dif.y + dif.z*dif.z);
	cameraChild->setTranslate(CVector3D(0, 0, len));

	CVector3D rot;
	rot.x = rot.z = CMath::PI/2; 
	rot.z += atan2(dif.y, dif.x);
	rot.x -= atan2(dif.z, len);

	cameraParent->setRotate(rot);
}

/* Place the camera somewhere making it point somewhere.
 */
void CKamera::moveAndPoint(float x1, float y1, float z1, float x2, float y2, float z2)
{
	moveAndPoint(CVector3D(x1, y1, z1), CVector3D(x2, y2, z2));
}

/* Place the camera somewhere making it point somewhere.
 */
void CKamera::moveAndPoint(const CVector3D &camera, const CVector3D &target)
{
	move(camera);
	point(target);
}

/* Return position of the camera 
 */
CVector3D CKamera::position(void)
{
	return pos;
}


/* Returns the position the camera is looking at
 */
CVector3D CKamera::pointing(void)
{
	return cameraParent->getTranslate();
}

/* Used to store the endpoints of a camera travel trayectory.
 * startPos - where the camera starts
 * startTarget - where the camera is looking at initially
 * startTime - acumulated time (usually 'now')
 * endPos - where the camera has to end
 * endTarget - where the camera should end looking at
 * endTime - 'now' + some seconds
 * endReturnValue - travelTrayectory returns this when the 
 * acumulated time is bigger than endTime. Should be nonzero.
 */
void CKamera::computeTrayectory(const CVector3D &startPos, const CVector3D &startTarget, float startTime, const CVector3D &endPos, const CVector3D &endTarget, float endTime, int endReturnValue)
{
	assert(endReturnValue);

	this->startPos = startPos;
	this->startTarget = startTarget;
	this->startTime = startTime;
	this->endPos = endPos;
	this->endTarget = endTarget;
	this->endTime = endTime;
	this->endReturnValue = endReturnValue;
}

/* Used to compute a trayectory using the actual camera state
 * as the beginning points.
 * startTime - acumulated time (usually 'now')
 * endPos - where the camera has to end
 * endTarget - where the camera should end looking at
 * endTime - 'now' + some seconds
 * endReturnValue - travelTrayectory returns this when the 
 * acumulated time is bigger than endTime. Should be nonzero.
 */
void CKamera::computeTrayectory(float startTime, const CVector3D &endPos, const CVector3D &endTarget, float endTime, int endReturnValue)
{
	computeTrayectory(pos, cameraParent->getTranslate(),
		startTime, endPos, endTarget, endTime, endReturnValue);
}		

/* Places the kamera in a position between the start and end 
 * values specified before with computeTrayectory. The time is
 * the accumulated time since the start of the program.
 * Returns: endReturnValue if time is bigger than the specified
 * endTime. Otherwise returns zero.
 */
int CKamera::travelTrayectory(float time)
{
	assert(endTime > startTime);
	assert(time >= 0);

	time = (time - startTime) / (endTime - startTime);
	if(time > 1) { // did we end? make sure we reach destination
		move(endPos);
		point(endTarget);
		return endReturnValue;
	}

	move(interpolateVector(time, startPos, endPos));
	point(interpolateVector(time, startTarget, endTarget));
	return 0;
}

/* Usually you want a camera which is behind the player and 
 * faces the other player. This computes a camera which will be
 * behind player1 looking at player2, slightly elevated over 
 * both. The result will be put in 'position' and 'target'.
 */
void CKamera::putCameraBehindPlayer(const CVector3D &karro1, const CVector3D &karro2, CVector3D &position, CVector3D &target)
{
	CVector3D dif = karro2 - karro1;
	position = karro1;
	position.z += 4 - dif.z * 0.1f;
	position.x -= dif.x * 0.1f;
	position.y -= dif.y * 0.1f;
	target = karro2;
	target.z -= 10;
}

/* Returns an interpolated position between two vectors, 
 * depending on the value of pos, which has to be 
 * inbetween 0 and 1.
 */
CVector3D CKamera::interpolateVector(float pos, const CVector3D &p1, const CVector3D &p2)
{
	CVector3D n;
	n.x = p1.x + (p2.x - p1.x) * pos;
	n.y = p1.y + (p2.y - p1.y) * pos;
	n.z = p1.z + (p2.z - p1.z) * pos;
	return n;
}

