/*  Katapult, version 1.
    Copyright (C) 2002 Grzegorz Adam Hankiewicz <gradha@terra.es> and
    Jon Lpez de Dicastillo <darwater@hotmail.com>.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "Katapult.h"
#include "Earth.h"
#include "Kamera.h"
#include "Karro.h"

#include "E3D_CInputDeviceKeyboardW32.h"
#include "E3D_CInputDeviceMouseW32.h"
#include "E3D_CFactoryFile.h"
#include "E3D_IReader.h"
#include "E3D_CRenderWOGL.h"
#include "E3D_CTransform.h"
#include "E3D_CMesh.h"
#include "E3D_CCamera.h"
#include "E3D_CMath.h"

#include <assert.h>

// stupid constructors.
CApplication* CApplication::create() { return new Katapult(); }
Katapult::Katapult(): render(NULL) {}

// destructor
bool Katapult::end()
{
	delete kamera;
	delete render;
	return true;
}

bool Katapult::eventResize(int width, int height)
{
	if (render != NULL)
		render->setViewport(0, 0, width, height);

	return true;
}

/* Initialization of the engine and the world objects.
 */
bool Katapult::init()
{
	render = NULL;

	// Init the input
	if (!input.addInputDevice(new CInputDeviceKeyboardW32())) return false;
	if (!input.addInputDevice(new CInputDeviceMouseW32())) return false;
	input.addInputListener(this);

	// Init the render
	render = new CRenderWOGL();

	if (render == NULL) return false;
	if (!render->init(window, 640, 480, false)) return false;

	CTransform* cylinder = new CTransform();
	cylinder->setName("Cylinder");
	cylinder->setRenderBoundingBox(true);
	scene.addNode(cylinder);
	scene.addChild(cylinder);

	// Create kamera
	kamera = new CKamera(scene);

	// Create world plane
	earth = new CEarth(scene, WORLD_WIDTH, WORLD_HEIGHT);

	// Create players
	karro1 = new CKarro(scene, "Karro1", 0);
	karro2 = new CKarro(scene, "Karro2", 1);

	CVector3D t = earth->zone(0);
	t.x += earth->zoneWidth() * 0.25f;
	t.y += earth->zoneHeight() * 0.5f;
	karro1->move(t);
	t = earth->zone(1, 1);
	t.x -= earth->zoneWidth() * 0.25f;
	t.y -= earth->zoneHeight() * 0.5f;
	karro2->move(t);

	// Sets the initial kamera travel trayectory
	timer.update();
	logicState = CAMERA_MOVE;

	CVector3D object(earth->width()/2, earth->height()/2, 0);
	kamera->computeTrayectory(
		CVector3D(object.x-0.1, object.y, 400), object,
		timer.getAccumTime(), 
		CVector3D(object.x-0.1, object.y, 20), object,
		timer.getAccumTime()+3, CAMERA_TO_PLAYER);

	return true;
}

/* This "runs" the game.
 */
bool Katapult::update()
{
	timer.update();
	input.update(0);
	updateLogic();
	render->startFrame();
	scene.update(0);
	scene.render(render);
	render->flip();
	render->endFrame();
	return true;
}

float alt1 = 6, alt2 = 1.5f;

/* Performs the logic of the real game.
 */
void Katapult::updateLogic(void)
{
	int n;
	switch (logicState) {
		case CAMERA_MOVE:
			n = kamera->travelTrayectory(timer.getAccumTime());
			if(n) logicState = n;
			return;
		case CAMERA_TO_PLAYER: logicCameraToPlayer(); return;
		case PLAYER_MOVE: logicPlayerMove(); return;
		case PLAYER_AIM: logicPlayerAim(); return;
		case FOLLOW_BULLET: logicFollowBullet(); return;
			return;
	}

	assert(0);
}

/* Moves the bullet through the air and the camera follows it
 */
void Katapult::logicFollowBullet(void)
{
	int freeze;
	CBoundingBox caja_karro;
	CBoundingBox caja_bala;

	CVector3D posBala;
	int n = karro1->getRecorrido(timer.getAccumTime(), posBala, freeze);
	CVector3D n1, n2;
	kamera->putCameraBehindPlayer(posBala, karro2->position(), n1, n2);
	n2.z += 11;
	n1.z -= 2;

	if(!freeze)		lastCameraPosition = n1;
	else			n1 = lastCameraPosition;

	kamera->moveAndPoint(n1, n2);
	if(n) {
		logicState = n;
		// This switches players...
		CKarro *k = karro2;
		karro2 = karro1;
		karro1 = k;

		caja_karro = karro1->getBoundingBox();
		caja_bala = karro2->getBalaBoundingBox();
		if(boundingBoxCollision(caja_karro, caja_bala))
			exit(0);

		logicState = CAMERA_MOVE;
		CVector3D n1, n2;
		kamera->putCameraBehindPlayer(karro1->position(),
			karro2->position(), n1, n2);

		kamera->computeTrayectory(timer.getAccumTime(), n1, n2,
		timer.getAccumTime() + 2, PLAYER_MOVE);
	}
}

/* Returns true is the bounding boxes overlap.
 */
bool Katapult::boundingBoxCollision(const CBoundingBox &p1, const CBoundingBox &p2)
{
	if(p1.max.x < p2.min.x) return false;
	if(p1.max.y < p2.min.y) return false;
	if(p1.max.z < p2.min.z) return false;
	if(p1.min.x > p2.max.x) return false;
	if(p1.min.y > p2.max.y) return false;
	if(p1.min.z > p2.max.z) return false;

	return true;
}

/* Handles creating the camera transition to one player.
 */
void Katapult::logicCameraToPlayer(void)
{
	CVector3D n1, n2;
	kamera->putCameraBehindPlayer(karro1->position(), 
		karro2->position(), n1, n2);
	logicState = CAMERA_MOVE;
	kamera->computeTrayectory(timer.getAccumTime(), n1, n2,
		timer.getAccumTime() + 4, PLAYER_MOVE);
}

/* This rotates the katapult to aim somewhere
 */
void Katapult::logicPlayerAim(void)
{
	float ang = karro1->angle();
	ang -= moveX * timer.getElapsedTime() * (CMath::PI / 3);
	karro1->angle(ang);

	CVector3D dest, pos;
	putCameraBehindPala(pos, dest);
	kamera->moveAndPoint(pos, dest);

	if(modeChange) {
		logicState = CAMERA_MOVE;
		CVector3D n1, n2;
		kamera->putCameraBehindPlayer(karro1->position(),
			karro2->position(), n1, n2);

		kamera->computeTrayectory(timer.getAccumTime(), n1, n2,
		timer.getAccumTime() + 1, PLAYER_MOVE);
	}
	if(fire) {
		// Insert here code for shooting
		karro1->disparar(70, timer.getAccumTime()+1, timer.getAccumTime() + 6, PLAYER_MOVE);
		logicState = FOLLOW_BULLET;
	}
}

/* Handles the movement of the katapult.
 */
void Katapult::logicPlayerMove(void)
{
	// This gets calculated once and correctly.
	static const float factor = earth->zoneWidth() / 4;

	// Revert direction if the player is really karro2
	CVector3D pos = karro1->position();
	if(karro1->zone()) {
		pos.x -= moveX * timer.getElapsedTime() * factor;
		pos.y -= moveY * timer.getElapsedTime() * factor;
	} 
	else {
		pos.x += moveX * timer.getElapsedTime() * factor;
		pos.y += moveY * timer.getElapsedTime() * factor;
	}

	// Clip new position to zone.
	CVector3D limit = earth->zone(karro1->zone(), 0);
	pos.x = MAX(limit.x, pos.x);
	pos.y = MAX(limit.y, pos.y);
	
	limit = earth->zone(karro1->zone(), 1);
	pos.x = MIN(limit.x, pos.x);
	pos.y = MIN(limit.y, pos.y);
	karro1->move(pos);

	// replace the camera
	CVector3D n1, n2;
	kamera->putCameraBehindPlayer(karro1->position(),
		karro2->position(),n1, n2);
	kamera->moveAndPoint(n1, n2);

	if(modeChange) {
		logicState = CAMERA_MOVE;
		CVector3D dest, pos;
		putCameraBehindPala(pos, dest);

		kamera->computeTrayectory(timer.getAccumTime(), 
			pos, dest, timer.getAccumTime() + 1, PLAYER_AIM);
	}
}

/*
 */
void Katapult::putCameraBehindPala(CVector3D &pos, CVector3D &dest)
{
	dest = karro1->position();
	pos = karro1->getPosBala();
	dest.z = pos.z;
	pos.z += 0.5f;
}

/* Called when the user interacts.
 */
void Katapult::inputChanged(CInputEvent& ie)
{
	CInputData* iData = ie.getInputData();
	CInputButton* button;

#define TECLA(X) \
	button = iData->getInputButton(CInputButton::X);\
	assert(button);\
	if(button->getPressed())

	TECLA(BUTTON_NAME_Q)
		exit(0);

	moveX = moveY = modeChange = fire = 0;

	TECLA(BUTTON_NAME_UP)				moveY++;
	TECLA(BUTTON_NAME_DOWN)				moveY--;
	TECLA(BUTTON_NAME_LEFT)				moveX--;
	TECLA(BUTTON_NAME_RIGHT)			moveX++;
	TECLA(BUTTON_NAME_M)				modeChange++;
	TECLA(BUTTON_NAME_RETURN)			modeChange++;
	TECLA(BUTTON_NAME_SPACE)			fire++;
}

